"""Collate an assignment and answer Jupyter notebooks into a single notebook.

This script is designed to support active reading.  It takes as input
a set of Jupyter notebooks as well as some target cells which define a set
of reading exercises.  The script processes the collection of notebooks
and builds a notebook which summarizes the responses to each question.

Based on work by Paul Ruvolo.
Rewritten by Oliver Steele
"""

import re
# from collections import Iterable, OrderedDict
from collections import namedtuple
from difflib import SequenceMatcher
from itertools import starmap

import nbformat

# QUESTION_RE = r'#+ (Exercise|Question)'


def nb_clear_outputs(nb):
    "Clear the output cells from a Jupyter notebook."
    for cell in nb.cells:
        if 'outputs' in cell:
            cell['outputs'] = []


def nbcollate(assignment_nb, submission_nbs, *, ids=None, labels=None, clear_outputs=False):
    """Create a notebook based on assignment_nb, that incorporates answers from student_nbs.

    Args:
        assignment_nb: a Jupyter notebook with the assignment
        submission_nbs: a dict or iterable whose values are notebooks with answers

    Returns:
        A Jupyter notebook
    """
    if isinstance(submission_nbs, dict):
        assert not ids
        ids = list(submission_nbs.keys())
        submission_nbs = list(submission_nbs.values())

    def label_cell(s_name):
        return nbformat.v4.new_markdown_cell(source='**{}**'.format(s_name))
    Opcode = namedtuple('opcode', ['op', 'i1', 'i2', 'j1', 'j2'])
    changes = sorted((oc.i2, i, oc, nb.cells[oc.j1:oc.j2])
                     for i, nb in enumerate(submission_nbs)
                     for oc in starmap(Opcode, NotebookMatcher(assignment_nb, nb).get_opcodes()))
    output_cells = assignment_nb.cells[:]
    di = 0
    for _, i, opcode, b_cells in changes:
        op, i1, i2, j1, j2 = opcode
        if op in ('insert', 'replace'):
            i0 = i2 + di
            b_cells = [c.copy() for c in b_cells]
            if ids:
                for c in b_cells:
                    c.metadata = c.metadata.copy()
                    c.metadata.nbcollate_source = ids[i]
            if labels:
                b_cells = [label_cell(labels[i])] + b_cells
            output_cells[i0:i0] = b_cells
            di += len(b_cells)
    nb = assignment_nb.copy()
    nb.cells = [c.copy() for c in output_cells if c.source.strip()]
    if clear_outputs:
        nb_clear_outputs(nb)
    return nb


def cell_strings(nb):
    return [cell.source.strip() for cell in nb.cells]


def NotebookMatcher(nb1, nb2):
    return SequenceMatcher(None, cell_strings(nb1), cell_strings(nb2))


def isections(nb):
    section = (None, [])
    for cell in nb.cells:
        m = re.match(r'^##+\s*(.+)', cell.source)
        if m:
            if section[-1]:
                yield section
            section = (m.group(1), [])
        section[-1].append(cell)
    if section[-1]:
        yield section


def remove_duplicate_answers(nb):
    dups = []
    for _, cells in isections(nb):
        seen = set()
        for c in cells:
            h = c.source.strip()
            if h in seen:
                dups.append(c)
            seen.add(h)
    for d in dups:
        nb.cells.remove(d)


def sort_answers(nb):
    dups = []
    out = []
    for _, cells in isections(nb):
        out += sorted(cells, key=lambda c: len(c.source.splitlines()))
    nb.cells = out


def get_cell_source_id(cell):
    return getattr(cell.metadata, 'nbcollate_source', None)


def get_answer_tuples(nb):
    "Return a set of tuples (student_id, prompt_title) of answered prompts."
    return {(title, get_cell_source_id(c))
            for title, cells in isections(nb)
            for c in cells if get_cell_source_id(c) is not None}



#!/usr/bin/env python
"""Create a combined notebook. The first path is the assignment notebook.
Remaining paths are student notebooks.
"""

import argparse
import logging
import os
import sys

import nbformat
import nbformat.reader
import nbcollate as nbc
from . import nbcollate


def safe_read(nbf):
    """A wrapper for nbformat.read, that prints a warning and returns None on
    bad notebooks.
    """
    try:
        return nbformat.read(nbf, as_version=4)
    except nbformat.reader.NotJSONError:
        print('while reading', nbf)


def capitalize(s):
    return s[:1].upper() + s[1:]


def collate(master_nb_path, submission_paths, args):
    """Collate notebooks.

    Arguments
    ---------
    master_nb_path: str
        The master notebook.
    submission_paths: [str]
        A list of notebook file pathnames.
    """
    if args.verbose:
        logging.basicConfig(format='%(message)s', level=logging.INFO)
    submission_nbs = [safe_read(nbf) for nbf in submission_paths]
    submission_nbs = [collated_nb for collated_nb in submission_nbs if collated_nb]
    master_nb = safe_read(master_nb_path)
    assert master_nb
    labels = None
    if args.label:
        labels = [capitalize(os.path.splitext(os.path.split(f)[1])[0].replace('-', ' '))
                  for f in submission_paths]
    collated_nb = nbcollate(master_nb, submission_nbs, labels=labels)
    suffix = "-collated"
    root, ext = os.path.splitext(args.notebook_files[0])
    collated_nb_path = "{}{}{}".format(root, suffix, ext)
    if args.out:
        collated_nb_path = os.path.join(
            args.out, os.path.split(collated_nb_path)[1])
    if not args.force and os.path.exists(collated_nb_path):
        # FIXME raise condition; instead open w/ os.O_CREAT | os.O_WRONLY
        err = FileExistsError()
        err.filename = collated_nb_path
        raise err
    if not args.dry_run:
        with open(collated_nb_path, 'w') as f:
            nbformat.write(collated_nb, f)
    print('wrote', collated_nb_path)


def main(args=sys.argv[1:]):
    "Create a collated notebook."
    parser = argparse.ArgumentParser(description=__doc__)
    nb_nargs = '*' if '--version' in args else '+'
    parser.add_argument('-f', '--force', action='store_true', help="Force overwrite existing file")
    parser.add_argument('-n', '--dry-run', help="Dry run")
    parser.add_argument('-o', '--out', type=str, help="Output directory")
    parser.add_argument('-v', '--verbose', action='store_true')
    parser.add_argument('--label', action='store_true', help="Label answers by notebook")
    parser.add_argument('--version', action='store_true')
    parser.add_argument('notebook_files', nargs=nb_nargs, metavar='NOTEBOOK_FILE')
    args = parser.parse_args(args)
    if args.version:
        print('nbcollate version', nbc.__version__)
        return
    if not args.notebook_files:
        parser.error('the following arguments are required: NOTEBOOK_FILE')
    master_file, *submission_files = args.notebook_files
    # Remove the master file from the answer files. This allows the CLI
    # to be used in the pattern `nbcollate master.ipynb *.ipynb`.
    if master_file in submission_files:
        submission_files = [f for f in submission_files if f != master_file]
    try:
        collate(master_file, submission_files, args)
    except FileExistsError:
        sys.stderr.write("Output file already exists. Repeat with --force to replace it.\n")
        sys.exit(1)



from collections import OrderedDict

from helpers import maybe_write_notebook, nb_sections, read_notebook, section_contains_string
from nbcollate import nbcollate, get_answer_tuples, nb_clear_outputs
import nbcollate as nbc

ASSIGNMENT_NB = read_notebook('assignment')
SUBMISSION_NBS = OrderedDict(
    (student_name, read_notebook(student_name))
    for student_name in ['student-1', 'student-2', 'student-3', 'student-4'])


def test_collate_without_names():
    nb = nbcollate(ASSIGNMENT_NB, SUBMISSION_NBS, clear_outputs=True)
    maybe_write_notebook(nb, 'anonymous.ipynb')

    assert nb.metadata
    assert nb.metadata == ASSIGNMENT_NB.metadata

    sections = nb_sections(nb)

    assert "Question 1" in sections
    assert section_contains_string(sections, "Question 1",
                                   'print("Student 1 answers question 1.")')
    assert section_contains_string(sections, "Question 1",
                                   'print("Student 2 answers question 1.")')
    # assert section_contains_string(sections, "Question 1", 'print("Student 2 answers question 1 with more code.")')
    assert section_contains_string(sections, "Question 1",
                                   'print("Student 3 answers question 1.")')

    assert "Question 2" in sections
    assert section_contains_string(sections, "Question 2",
                                   'print("Student 1 answers question 2.")')
    assert not section_contains_string(
        sections, "Question 2", 'print("Student 2 answers question 2.")')
    assert section_contains_string(sections, "Question 2",
                                   'print("Student 3 answers question 2.")')

    assert "A Quick Poll" in sections
    assert section_contains_string(sections, "A Quick Poll",
                                   'Student 1 answers the quick poll')
    assert section_contains_string(sections, "A Quick Poll",
                                   'Student 2 answers the quick poll')
    assert not section_contains_string(sections, "A Quick Poll",
                                       'Student 3 answers the quick poll')


def assert_golden(nb, basename, clear_outputs=False):
    gm = read_notebook(basename)
    if clear_outputs:
        nb_clear_outputs(gm)
    maybe_write_notebook(nb, basename)
    assert nb == gm


def test_nbcollate_without_labels():
    nb = nbcollate(ASSIGNMENT_NB, SUBMISSION_NBS.values(), clear_outputs=True)
    assert_golden(nb, 'assignment-collated.ipynb', clear_outputs=True)

    nb = nbcollate(ASSIGNMENT_NB, SUBMISSION_NBS.values(), clear_outputs=False)
    assert_golden(nb, 'assignment-collated.ipynb', clear_outputs=False)


def test_nbcollate_sort():
    submission_nbs = [read_notebook('sorted-%d.ipynb' % i) for i in [1, 2]]
    nb = nbcollate(ASSIGNMENT_NB, submission_nbs, clear_outputs=True)
    nbc.sort_answers(nb)
    assert_golden(nb, 'sorted.ipynb', clear_outputs=True)


def test_nbcollate_remove_dups():
    submission_nbs = [read_notebook('dups-%d.ipynb' % i) for i in [1, 2]]
    nb = nbcollate(ASSIGNMENT_NB, submission_nbs, clear_outputs=True)
    nbc.remove_duplicate_answers(nb)
    assert_golden(nb, 'deduped.ipynb', clear_outputs=True)


def test_report_missing_answers():
    answers = get_answer_tuples(nbcollate(ASSIGNMENT_NB, SUBMISSION_NBS))
    assert {title for title, _ in answers} == {'A Quick Poll', 'Question 1', 'Question 2'}
    assert {student for _, student in answers} == {'student-1', 'student-2', 'student-3', 'student-4'}

    assert ('Question 1', 'student-1') in answers
    assert ('Question 1', 'student-2') in answers
    assert ('Question 1', 'student-3') in answers

    assert ('Question 2', 'student-1') in answers
    assert ('Question 2', 'student-2') not in answers
    assert ('Question 2', 'student-3') in answers



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['nbcollate']

package_data = \
{'': ['*']}

install_requires = \
['nbformat~=4.4.0']

setup(name='nbcollate',
      version='0.3.1',
      description='Collate Jupyter classroom assignment and submission notebooks',
      author='Oliver Steele',
      author_email='steele@osteele.com',
      url='https://github.com/osteele/nbcollate',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      python_requires='>=3.5',
     )



import itertools
import os
import re
from collections import OrderedDict, namedtuple

import nbformat


def read_notebook(basename):
    "Read notebook `basename` from the test files directory."
    if not basename.endswith('.ipynb'):
        basename += '.ipynb'
    return nbformat.read(
        os.path.join(os.path.dirname(__file__), 'files', basename),
        as_version=4)


def maybe_write_notebook(nb, basename):
    "If PYTEST_SAVE_OUTPUTS is set, save the notebook to test/build/{basename}.ipynb."
    if os.environ.get('PYTEST_SAVE_OUTPUTS', False) in {False, "0", "false"}:
        return
    if not basename.endswith('.ipynb'):
        basename += '.ipynb'
    output_path = os.path.join(os.path.dirname(__file__), 'build', basename)
    os.makedirs(os.path.dirname(output_path), exist_ok=True)
    nbformat.write(nb, open(output_path, 'w'))


class NotebookSection(namedtuple('_NotebookSection', 'title children cells')):

    def __str_lines__(self):
        prefix = '  '
        return itertools.chain(['Section: ' + self.title], (
            prefix + 'cell: ' + repr(cell.source)
            for cell in self.cells), (prefix + line
                                      for child in self.children.values()
                                      for line in child.__str_lines__()))

    def __str__(self):
        return '\n'.join(self.__str_lines__())

    def __contains__(self, title):
        return title in self.children


def nb_sections(nb):
    """Return a tree that represents the sections (by header) of a notebook.

    A Section is a tuple of a (dict, cells), where dict is an OrderedDict
    that maps titles to Sections.
    """

    def next_key(section, title):
        key = title
        if key in parent:
            candidates = map(lambda n: ("%s (%d)" % (key, n)).strip(),
                             itertools.count(1))
            key = next(itertools.filterfalse(parent.__contains__, candidates))
        return key

    path = []
    # import pudb; pu.db
    for cell in nb.cells:
        if cell.cell_type == 'markdown' and cell.source.startswith('#'):
            m = re.match(r'(#+)\s*(.*)', cell.source)
            assert m
            hash_prefix, title = m.groups()
            level = len(hash_prefix)

            while level < len(path) + 1:
                current_section = path.pop()
            while level > len(path):
                current_section = NotebookSection(title
                                                  if level == len(path) + 1
                                                  else '', OrderedDict(), [])
                if path:
                    parent = path[-1]
                    key = next_key(parent, title
                                   if level == len(path) + 1 else '')
                    parent.children[key] = current_section
                path.append(current_section)

        current_section.cells.append(cell)

    return path[0]


# TODO make this a matcher or something
def section_contains_string(nbtree, section_title, text):
    return any(text in cell.source
               for cell in nbtree.children[section_title].cells)


if __name__ == '__main__':
    assignment_nb = read_notebook('assignment')
    sections = nb_sections(assignment_nb)
    print(sections)
